#region

using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;
using JetBrains.Annotations;
using MeshVR;
using MVR.FileManagementSecure;
using SimpleJSON;
using UnityEngine;
using UnityEngine.Events;
using UnityEngine.UI;
using JSONClass = SimpleJSON.JSONClass;

#endregion

namespace SceneNinja {
    public class SceneNinja : MVRScript {

        private const string _settingsDir = "Saves/PluginData/Venkman/SceneNinja";
        private const string _settingsPath = "Saves/PluginData/Venkman/SceneNinja/SceneNinjaDefault.json";
        private const string _ignoreScenesPath = "Saves/PluginData/Venkman/SceneNinja/IgnoreScenes.json";
        private PlayerData[] _players;
        private List<string> _scenesToIgnore = new List<string>();
        private List<JSONStorableBool> _ignoreBools = new List<JSONStorableBool>();
        private List<JSONStorableString> _ignoreStrings = new List<JSONStorableString>();
        private const bool _debug = true;
        // private UIDynamic _lspacer;
        // private UIDynamic _rspacer;
        // private UIDynamicButton _addButton;
        // private string _currentScene = "";

        public override void Init() {
            _players = new PlayerData[4];
            _scenesToIgnore = new List<string>();
            // Example storable; you can also create string, float and action JSON storables
            var players = new List<PlayerData>();
            var fileExists = FileManagerSecure.FileExists(_settingsPath);
            if (fileExists) {
                var raw = FileManagerSecure.ReadAllText(_settingsPath);
                var data = JSON.Parse(raw).AsObject;
                foreach (KeyValuePair<string,JSONNode> obj in data.AsObject) {
                    try {
                        LogMessage("Loading player: " + obj.Value["id"]);
                        var player = new PlayerData(obj.Value);
                        player.LoadVars();
                        players.Add(player);
                    } catch (Exception e) {
                        LogMessage("Exception Loading Player Data: " + e.Message + " at " + e.StackTrace);
                    }
                }
            } else {
                SuperController.LogMessage("Creating first-time default config.");
                var p1 = new PlayerData("F1");
                var p2 = new PlayerData("M1");
                var p3 = new PlayerData("F2");
                var p4 = new PlayerData("M2");
                players.Add(p1);
                players.Add(p2);
                players.Add(p3);
                players.Add(p4);
                _players = players.ToArray();
                foreach(var player in _players) player.LoadVars();
                SuperController.LogMessage("Saving config, first load.");

                SaveConfig();
            }

            _players = players.ToArray();
            foreach (var atom in _players) {
                atom.Initialize(this);
            }

            var ignoreExists = FileManagerSecure.FileExists(_ignoreScenesPath);
            if (ignoreExists) {
                var toIgnore = FileManagerSecure.ReadAllText(_ignoreScenesPath);
                var ignoreSplit = toIgnore.Split(Char.Parse("\n"));
                foreach (var ignore in ignoreSplit) {
                    if (ignore.Replace(" ", "") == "") continue;
                    var descTitle = ignore.Replace(" ", "");
                    var jsb = new JSONStorableBool(descTitle, false, RemoveIgnore);
                    RegisterBool(jsb);
                    var staticDescString = new JSONStorableString(descTitle, ignore);
                    var staticDescStringField = CreateTextField(staticDescString);
                    RegisterString(staticDescString);
                    staticDescStringField.backgroundColor = new Color(1f, 1f, 1f, 0f);
                    disableScrollOnText(staticDescStringField);
                    CreateUIToggle(jsb,true,"Remove");
                    _ignoreBools.Add(jsb);
                    _ignoreStrings.Add(staticDescString);
                }
            }
            // _lspacer = CreateSpacer();
            // _rspacer = CreateSpacer(true);
            // _addButton = CreateUIButton("Add current scene to ignore.",false, AddCurrentScene);
            OnSceneLoaded();
        }

        // private void AddCurrentScene() {
        //     RemoveSpacer(_lspacer);
        //     RemoveSpacer(_rspacer);
        //     RemoveButton(_addButton);
        //     var ignore = GetScene();
        //     if (string.IsNullOrEmpty(ignore)) {
        //         SuperController.LogMessage("Can't add current scene, you need to save it first.");
        //         return;
        //     }
        //     if (!_scenesToIgnore.Contains(ignore)) {
        //         _scenesToIgnore.Add(ignore);
        //     }
        //     var descTitle = ignore.Replace(" ", "");
        //     var jsb = new JSONStorableBool(descTitle, false, RemoveIgnore);
        //     RegisterBool(jsb);
        //     var staticDescString = new JSONStorableString(descTitle, ignore);
        //     var staticDescStringField = CreateTextField(staticDescString);
        //     RegisterString(staticDescString);
        //     staticDescStringField.backgroundColor = new Color(1f, 1f, 1f, 0f);
        //     disableScrollOnText(staticDescStringField);
        //     CreateUIToggle(jsb,true,"Remove");
        //     _ignoreBools.Add(jsb);
        //     _ignoreStrings.Add(staticDescString);
        //     _lspacer = CreateSpacer();
        //     _rspacer = CreateSpacer(true);
        //     _addButton = CreateUIButton("Add current scene to ignore.",false, AddCurrentScene);
        // }

        private void RemoveIgnore(JSONStorableBool jb) {
            var toRemove = jb.name;
            var newList = new List<string>();
            var newIb = new List<JSONStorableBool>();
            var newIs = new List<JSONStorableString>();
            foreach (var ignore in _scenesToIgnore) {
                var desc = ignore.Replace(" ", "");
                if (ignore != toRemove) {
                    newList.Add(ignore);
                }

                foreach (var b in _ignoreBools) {
                    if (b.name == desc) {
                        RemoveToggle(b);
                        DeregisterBool(b);
                    } else {
                        newIb.Add(b);
                    }
                }

                _ignoreBools = newIb;

                foreach (var s in _ignoreStrings) {
                    if (s.name == desc) {
                        RemoveTextField(s);
                        DeregisterString(s);
                    } else {
                        newIs.Add(s);
                    }
                }

                _ignoreStrings = newIs;
            }
            _scenesToIgnore = newList;
            SaveIgnore();
        }

        private void SaveIgnore() {
            var saveString = "";
            for (var i = 0; i < _scenesToIgnore.Count; i++) {
                saveString += _scenesToIgnore[i];
                if (i != _scenesToIgnore.Count - 1) {
                    saveString += "\n";
                }
            }
            FileManagerSecure.WriteAllText(_ignoreScenesPath, saveString);
        }


        // called first
        private void OnEnable() {
            SuperController.singleton.onSceneLoadedHandlers += OnSceneLoaded;
        }

        // ReSharper disable Unity.PerformanceAnalysis
        private IEnumerator WaitForLoad() {
            var then = DateTime.Now;
            while (true) {
                LogMessage("Waiting for load to complete...");
                var now = DateTime.Now;
                if (now - then > TimeSpan.FromSeconds(10)) {
                    if (!SuperController.singleton.isLoading) break;
                    yield return null;
                }
                if (!SuperController.singleton.isLoading) break;
                yield return null;
            }
            LogMessage("Loading!");
            var atoms = SuperController.singleton.GetAtoms();
            var guys = new List<Atom>();
            var gals = new List<Atom>();
            foreach (var atom in atoms.Where(atom => atom.type == "Person")) {
                if (IsMale(atom)) {
                    guys.Add(atom);
                } else {
                    gals.Add(atom);
                }
            }

            var closestMale = GetClosestPerson(guys);
            var closestFemale = GetClosestPerson(gals);
            //var asf = new AsyncFlag("SceneNinja");
            //SuperController.singleton.ResetSimulation(asf, true);

            if (closestMale != null) {
                foreach (var person in guys) {
                    var playerData = _players[3];
                    if (person.name == closestMale.name) {
                        playerData = _players[1];
                    }

                    ApplyAppearance(person, playerData);
                    ApplyPlugins(person, playerData);
                    ApplyAppearance(person, playerData);
                }
            }

            if (closestFemale != null) {
                foreach (var person in gals) {
                    var playerData = _players[2];
                    if (person.name == closestFemale.name) {
                        playerData = _players[0];
                    }

                    ApplyAppearance(person, playerData);
                    ApplyPlugins(person, playerData);
                    ApplyAppearance(person, playerData);
                }
            }

            if (SuperController.singleton.freezeAnimation) {
                LogMessage("Unfreezing!");
                SuperController.singleton.SetFreezeAnimation(false);
            }

            //asf.Raise();
        }

        private void OnSceneLoaded() {
            // var ignore = GetScene();
            // if (string.IsNullOrEmpty(ignore)) {
            //     SuperController.LogMessage("Can't check current scene, you need to save it first.");
            // }
            // if (_scenesToIgnore.Any(toIgnore => toIgnore == ignore)) {
            //     SuperController.LogMessage("Ignoring load for scene: " + ignore);
            //     return;
            // }
            StartCoroutine(WaitForLoad());
        }

        private void ApplyPlugins(Atom target, PlayerData playerData) {
            var presetFilePath = playerData.PluginsPath;
            var normalPath = presetFilePath == "" ? "" : SuperController.singleton.NormalizePath(presetFilePath);
            var atom = SuperController.singleton.GetAtomByUid(target.name);
            manager = atom.GetStorableByID("PluginManager") as MVRPluginManager;
            var existingSettings = new JSONClass { { "setUnlistedParamsToDefault", "true" } };
            var exStorables = new JSONArray();
            var existing = new Dictionary<string, JSONNode>();
            var indices = new List<int>();


            if (manager != null) {
                var managerJson = manager.GetJSON();
                exStorables.Add("PluginManager", managerJson);
                var plugins = manager.GetJSON()["plugins"];
                foreach (KeyValuePair<string, JSONNode> f in plugins.AsObject) {
                    existing[f.Key] = f.Value;
                    var index = f.Key.Replace("plugin#", "");
                    var idx = int.Parse(index);
                    indices.Add(idx);
                }
            }

            foreach (var st in atom.GetStorableIDs()) {
                if (st.Contains("plugin#")) {
                    exStorables.Add(st, atom.GetStorableByID(st).GetJSON(true, true, true));
                }
            }

            existingSettings.Add("storables", exStorables);

            if (presetFilePath != "" && FileManagerSecure.FileExists(presetFilePath)) {
                var presetPackageName = "";
                if (presetFilePath.Contains(":")) {
                    presetPackageName = presetFilePath.Substring(0, presetFilePath.IndexOf(':'));
                }

                var data = FileManagerSecure.ReadAllText(normalPath);
                var presetJson = JSON.Parse(data);

                if (presetPackageName != "") {
                    var presetJsonString = presetJson.ToString();
                    if (presetJsonString.Contains("SELF:")) {
                        presetJsonString = presetJsonString.Replace("SELF:", presetPackageName + ":");
                        presetJson = JSON.Parse(presetJsonString);
                    }
                }

                var mappings = new Dictionary<string, JSONNode>();
                var plugs = presetJson["storables"][0]["plugins"];
                if (plugs == null) {
                    SuperController.LogMessage("Unable to parse plugin data, nothing to do: " + presetFilePath);
                    return;
                }
                foreach (KeyValuePair<string, JSONNode> storableJson in plugs.AsObject) {
                    mappings[storableJson.Key] = storableJson.Value;
                }

                var ignoreVersion = playerData.IgnorePluginVersion;
                var newIndexes = new Dictionary<string, string>();
                if (playerData.MergePlugins) {
                    LogMessage("Merging plugins.");
                    foreach (var map in mappings) {
                        var val = map.Value.ToString();
                        foreach (var ex in existing) {
                            var exVal = ex.Value.ToString();
                            if (ignoreVersion) {
                                var valPart = val.Split(char.Parse(":"));
                                var exPart = exVal.Split(char.Parse(":"));
                                var valParts = valPart[0].Split(char.Parse("."));
                                var exParts = exPart[0].Split(char.Parse("."));
                                valParts[valParts.Length - 1] = "latest";
                                exParts[exParts.Length - 1] = "latest";
                                exPart[0] = string.Join(".", exParts);
                                valPart[0] = string.Join(".", valParts);
                                var vVal = string.Join(":", valPart);
                                var vexVal = string.Join(":", exPart);
                                val = vVal;
                                exVal = vexVal;
                            }

                            if (val == exVal && map.Key != ex.Key) {
                                newIndexes[map.Key] = ex.Key.Replace("\"", "");
                                break;
                            }
                        }
                    }

                    foreach (var index in from map in mappings select map.Key into index where !newIndexes.ContainsKey(index) where existing.ContainsKey(index) select index) {
                        for (var i = 0; i < 1000; i++) {
                            if (!indices.Contains(i)) {
                                indices.Add(i);
                                newIndexes[index] = "plugin#" + i;
                                break;
                            }
                        }
                    }
                } else LogMessage("No merge.");

                var newObj = new JSONClass();
                foreach (KeyValuePair<string, JSONNode> node in presetJson.AsObject) {
                    if (node.Key == "storables") {
                        var storables = node.Value.AsArray;
                        var newStorables = new JSONArray();
                        for (var i = 0; i < storables.Count; i++) {
                            var storable = storables[i].AsObject;
                            var id = storable["id"].Value;
                            id = id.Replace("\"", "");
                            if (id.Contains("PluginManager")) {
                                var plugins = storable["plugins"];
                                var newPlugins = new JSONClass();
                                foreach (KeyValuePair<string, JSONNode> p in plugins.AsObject) {
                                    if (newIndexes.ContainsKey(p.Key)) {
                                        LogMessage("Reindexing " + p.Key + " to " + newIndexes[p.Key]);
                                        newPlugins.Add(newIndexes[p.Key], p.Value);
                                    } else {
                                        LogMessage("Adding " + p.Key + " " + p.Value);
                                        newPlugins.Add(p.Key, p.Value);
                                    }
                                }

                                storable["plugins"] = newPlugins;
                                newStorables.Add(storable);
                            } else {
                                var idx = id.Split(char.Parse("_"))[0];
                                if (newIndexes.ContainsKey(idx)) {
                                    var newIdx = newIndexes[idx];
                                    var newId = storable["id"].ToString().Replace(idx, newIdx);
                                    newId = newId.Replace("\"", "");
                                    storable["id"] = newId;
                                    LogMessage($"Setting storable index from {idx} to {newId}");
                                }

                                newStorables.Add(storable);
                            }
                        }

                        newObj.Add("storables", newStorables);
                    } else {
                        newObj.Add(node.Key, node.Value);
                    }
                }

                if (playerData.MergePlugins) {
                    var mergedStorables = new JSONArray();
                    var existingStorables = existingSettings["storables"].AsArray;
                    var newStorables = newObj["storables"].AsArray;
                    var newAdded = new List<string>();
                    var existingPlugins = new List<string>();
                    var plugMan = new JSONClass();
                    foreach (JSONClass exS in existingStorables) {
                        var exId = exS["id"].ToString().Replace("\"", "");
                        if (exId != "PluginManager") {
                            var replaced = false;
                            foreach (JSONClass newS in newStorables) {
                                var nuId = newS["id"].ToString().Replace("\"", "");
                                if (nuId == exId) {
                                    LogMessage($"No, really, replacing {exId} data.");
                                    replaced = true;
                                    mergedStorables.Add(newS);
                                    newAdded.Add(nuId);
                                    break;
                                }
                            }

                            if (!replaced) {
                                LogMessage("Adding storable data for " + exS["id"]);
                                mergedStorables.Add(exS);
                            }
                        } else {
                            plugMan = exS.AsObject;
                            var plugins = exS["plugins"];
                            existingPlugins.AddRange(plugins.AsObject.Keys);
                        }
                    }

                    foreach (JSONClass newS in newStorables) {
                        var nuId = newS["id"].ToString().Replace("\"", "");
                        if (nuId != "PluginManager") {
                            if (!newAdded.Contains(newS["id"])) {
                                mergedStorables.Add(newS);
                            }
                        } else {
                            var newPlugins = newS["plugins"];
                            var plugins = plugMan["plugins"];
                            foreach (KeyValuePair<string, JSONNode> np in newPlugins.AsObject) {
                                if (!existingPlugins.Contains(np.Key)) {
                                    LogMessage("Adding missing plugin: " + np.Key);
                                    plugins.Add(np.Key, np.Value);
                                }
                            }

                            plugMan["plugins"] = plugins;
                            var reOrdered = new JSONArray();
                            reOrdered.Add(plugMan.AsObject);
                            foreach (JSONClass ms in mergedStorables) {
                                reOrdered.Add(ms);
                            }

                            mergedStorables = reOrdered;
                        }
                    }

                    newObj["storables"] = mergedStorables;
                }

                var presetJs = atom.GetStorableByID("PluginPresets");
                atom.SetLastRestoredData(newObj);
                var pm = presetJs.GetComponentInChildren<PresetManager>();
                pm.LoadPresetFromJSON(newObj);
            }
        }

        private void ApplyAppearance(Atom atom, PlayerData playerData) {
            var tempPresetLockStore = new PresetLockStore();
            if (atom.type == "Person" && !playerData.MergeClothing) {
                tempPresetLockStore.StorePresetLocks(atom, true, playerData.SuppressClothing);
            }

            var presetPackageName = "";
            var presetFilePath = playerData.AppearancePath;
            if (presetFilePath == "") {
                return;
            }

            if (presetFilePath.Contains(":")) {
                presetPackageName = presetFilePath.Substring(0, presetFilePath.IndexOf(':'));
            }

            var presetJson =
                (JSONClass)SuperController.singleton.LoadJSON(SuperController.singleton.NormalizePath(presetFilePath));

            if (presetPackageName != "") {
                var presetJsonString = presetJson.ToString();
                if (presetJsonString.Contains("SELF:")) {
                    presetJsonString = presetJsonString.Replace("SELF:", presetPackageName + ":");
                    presetJson = (JSONClass)JSON.Parse(presetJsonString);
                }
            }

            var presetStorablesJson = (JSONArray)presetJson["storables"];

            var atomScale = atom.GetStorableByID("rescaleObject").GetFloatJSONParam("scale");
            if (playerData.SuppressScale) {
                foreach (JSONClass storableJson in presetStorablesJson) {
                    if (storableJson["id"].Value == "rescaleObject") {
                        storableJson["scale"].AsFloat = atomScale.val;
                    }
                }
            }

            var presetJs = atom.GetStorableByID("AppearancePresets");
            var pm = presetJs.GetComponentInChildren<PresetManager>();

            if (playerData.MergeClothing) {
                presetJson = MergeClothing(atom, presetJson);
            }

            atom.SetLastRestoredData(presetJson);
            pm.LoadPresetFromJSON(presetJson);
            if (atom.type == "Person" && playerData.SuppressClothing && !playerData.MergeClothing) {
                tempPresetLockStore.RestorePresetLocks(atom);
            }


        }

        private JSONClass MergeClothing(Atom atom, JSONClass preset) {
            var addClothing = new Dictionary<string,JSONClass>();
            var storables = preset.AsObject["storables"];
            JSONNode clothing = null;
            foreach (JSONClass storable in storables.AsArray) {
                if (storable["id"].Value == "geometry") {
                    clothing = storable["clothing"];
                    foreach (JSONClass garment in clothing.AsArray) {
                        addClothing[garment["id"].Value] = garment;
                    }
                    break;
                }
            }

            if (clothing != null) {
                var atomGeo = atom.GetStorableByID("geometry");
                foreach (KeyValuePair<string, JSONNode> atomStorable in atomGeo.GetJSON()) {
                    SuperController.LogMessage("Foo3: " + atomStorable.Key);
                    if (atomStorable.Key == "clothing") {
                        var exArray = atomStorable.Value.AsArray;
                        foreach (JSONClass toAdd in exArray) {
                            if (!addClothing.ContainsKey(toAdd["id"].Value)) {
                                SuperController.LogMessage("Adding clothing from atom to preset: " + toAdd["id"].Value);
                                clothing.AsArray.Add(toAdd);
                            }
                        }
                        break;
                    }
                }
            }

            for (var i = 0; i < storables.Count; i++) {
                var storable = storables[i];
                if (storable["id"].Value == "geometry") {
                    storable["clothing"] = clothing;
                    storables[i] = storable;
                }
            }

            return preset;
        }


        // called when the game is terminated
        private void OnDisable() {
            SuperController.singleton.onSceneLoadedHandlers -= OnSceneLoaded;
        }


        public UIDynamicButton CreateUIButton(string label, bool rightSide = false, UnityAction callback = null,
            float height = 50, string helpText = "") {
            var button = CreateButton(label, rightSide);
            button.height = height;
            if (callback != null) {
                button.button.onClick.AddListener(callback);
            }

            return button;
        }

        public void CreateUITextField(JSONStorableString label, bool rightSide = false,
            int height = 50, string helpText = "") {
            var field = CreateTextField(label, rightSide);
            var layout = field.GetComponent<LayoutElement>();
            layout.preferredHeight = layout.minHeight = height;
            field.height = height;
        }

        public void CreateUIToggle(JSONStorableBool label, bool rightSide = false, string toggleLabel = "",
            UnityAction<bool> callback = null,
            float height = 50, string helpText = "") {
            var toggle = CreateToggle(label, rightSide);
            toggle.height = height;
            if (!string.IsNullOrEmpty(toggleLabel)) {
                toggle.label = toggleLabel;
            }

            if (callback != null) {
                toggle.toggle.onValueChanged.AddListener(callback);
            }
        }


        private static void disableScrollOnText(UIDynamicTextField target) {
            var targetSr = target.UItext.transform.parent.transform.parent.transform.parent.GetComponent<ScrollRect>();
            if (targetSr != null) {
                targetSr.horizontal = false;
                targetSr.vertical = false;
            }
        }

        public static void LogMessage(string text) {
            if (_debug) {
                SuperController.LogMessage(text);
            }
        }

        public void CreateUILabel(string descTitle, bool rightSide, int fieldHeight = 40, string helpText = "") {
            var staticDescString = new JSONStorableString(descTitle.Replace(" ", ""), descTitle);
            var staticDescStringField = CreateTextField(staticDescString, rightSide);
            staticDescStringField.backgroundColor = new Color(1f, 1f, 1f, 0f);
            var layout = staticDescStringField.GetComponent<LayoutElement>();
            layout.preferredHeight = layout.minHeight = fieldHeight;
            staticDescStringField.height = fieldHeight;
            disableScrollOnText(staticDescStringField);
        }

        public void UpdatePlayerData(PlayerData playerData) {
            var i = 0;
            if (playerData.ID == "M1") i = 1;
            if (playerData.ID == "F2") i = 2;
            if (playerData.ID == "M2") i = 3;
            _players[i] = playerData;
            SuperController.LogMessage("Saving config, player updated!");
            SaveConfig();
        }

        // private string GetScene() {
        //     SuperController.LogMessage("Getting scene?");
        //     var foo2 = SuperController.singleton.GetSaveJSON();
        //     SuperController.LogMessage("No: " + foo2);
        //     var loadJson = SuperController.singleton.loadJson;
        //
        //     if (loadJson == null) {
        //         SuperController.LogMessage("Unable to find loadJSON.");
        //         return "";
        //     }
        //
        //     foreach (var foo in loadJson.AsObject) {
        //
        //     }
        //     if (SuperController.singleton.loadJson.ToString()==null) return "";
        //     return "scene";
        // }

        private void SaveConfig() {
            if (!FileManagerSecure.DirectoryExists(_settingsDir)) {
                FileManagerSecure.CreateDirectory(_settingsDir);
            }

            var newData = new JSONClass();
            foreach (var player in _players) {
                var jo = player.ToJson();
                newData.Add(player.ID, jo);
            }

            FileManagerSecure.WriteAllText(_settingsPath, newData.ToString(""));
        }

        private static bool IsMale(Atom atom) {
            var isMale = false;
            var femStrings = new[] { "female", "girl", "her", "woman" };
            var guyStrings = new[] { "male", "guy", "him", "man", "futa" };
            // If the person atom is not "On", then we cant determine their gender it seems as GetComponentInChildren<DAZCharacter> just returns null
            if (atom.on && (atom.containingSubScene == null || atom.containingSubScene.containingAtom.on) &&
                atom.type == "Person") {
                var targets = new[] { "FemaleAnatomy", "MaleAnatomy", "geometry" };
                foreach (var control in atom.GetStorableIDs()) {
                    if (targets.Contains(control)) {
                        var storable = atom.GetStorableByID(control).GetJSON();
                        if (control == "geometry") {
                            var morphs = storable["morphs"].AsArray;
                            var otherMorphs = storable["morphsOtherGender"].AsArray;
                            var useFmOnM = storable["useFemaleMorphsOnMale"];
                            var useMmOnF = storable["useMaleMorphsOnFeale"];
                            if (useFmOnM.AsBool) {
                                LogMessage("Female morphs on male!");
                                return true;
                            }
                            if (useMmOnF.AsBool) {
                                LogMessage("Male morphs on female!");
                                return false;
                            }
                            foreach (JSONClass morph in morphs) {
                                var name = morph["name"].ToString();
                                if (name.Contains("Figure ID")) {
                                    LogMessage("Figure id: " + name);
                                    if (name.ToLower().Contains("female")) {
                                        LogMessage("Morph name is female.");
                                        return false;
                                    }
                                    if (name.ToLower().Contains("male")) {
                                        LogMessage("Morph name is female.");
                                        return false;
                                    }
                                }

                                if (name.ToLower().Contains("penis")) {
                                    LogMessage("This player has a dick: " + name);
                                    return true;
                                }
                                if (name.ToLower().Contains("labia")) {
                                    LogMessage("This player has a vagina: " + name);
                                    return false;
                                }
                            }
                            foreach (JSONClass morph in otherMorphs) {
                                var name = morph["name"].ToString();
                                if (name.ToLower().Contains("penis")) {
                                    LogMessage("This player has a dick: " + name);
                                    return true;
                                }
                                if (name.ToLower().Contains("labia")) {
                                    LogMessage("This player has a vagina: " + name);
                                    return false;
                                }
                            }
                            var charSkin = storable["character"].Value;
                            LogMessage("Charskin: " + charSkin);
                            if (femStrings.Any(fem => charSkin.ToLower().Contains(fem))) {
                                LogMessage("Fem char string found, returning false.");
                                return false;
                            }
                            if (guyStrings.Any(guy => charSkin.ToLower().Contains(guy))) {
                                LogMessage("Guy char string found, returning true.");
                                return true;
                            }
                        } else {
                            LogMessage($"Anatomy flag for {control} is {storable["enabled"].Value}");
                            var active = storable["enabled"].AsBool;
                            if (active && control == "FemaleAnatomy") {
                                LogMessage("Enabled female anatomy flag set, ret false.");
                                return false;
                            }
                            if (active && control == "MaleAnatomy") {
                                LogMessage("Enabled male anatomy flag set, ret true.");
                                return true;
                            }
                        }

                    }
                }
                isMale = atom.GetComponentInChildren<DAZCharacter>().name.StartsWith("male") ||
                         atom.GetComponentInChildren<DAZCharacter>().name.StartsWith("jarlee") ||
                         atom.GetComponentInChildren<DAZCharacter>().name.StartsWith("jarjulian") ||
                         atom.GetComponentInChildren<DAZCharacter>().name.StartsWith("Futa");
            }

            return isMale;
        }

        [CanBeNull]
        private static Atom GetClosestPerson(List<Atom> atoms) {
            Atom closestPerson = null;
            var m1 = float.MaxValue;
            var isIndexed = true;
            var minIndex = int.MaxValue;
            var personIdx = new Dictionary<int, Atom>();
            foreach (var person in atoms) {
                if (person.hidden && !SuperController.singleton.showHiddenAtoms) {
                    continue;
                }

                var d2 = float.MaxValue;
                foreach (var rb in person.rigidbodies) {
                    var rbPosition = rb.transform.position;
                    var playerHeadTransform = CameraTarget.centerTarget.transform;
                    var distance = Vector3.Distance(rbPosition, playerHeadTransform.position);
                    var resultString = Regex.Match(person.name, @"\d+").Value;
                    try {
                        var idx = int.Parse(resultString);
                        personIdx[idx] = person;
                    } catch (Exception) {
                        isIndexed = false;
                    }

                    if (distance < d2) {
                        d2 = distance;
                    }
                }

                if (d2 < m1) {
                    closestPerson = person;
                }
            }

            if (isIndexed) {
                foreach (var pi in personIdx) {
                    if (pi.Key < minIndex) {
                        minIndex = pi.Key;
                        closestPerson = pi.Value;
                    }
                }
            }

            return closestPerson;
        }
    }

    public class PresetLockStore {
        private bool _generalPresetLock;
        private bool _appPresetLock;
        private bool _posePresetLock;
        private bool _animationPresetLock;
        private bool _glutePhysPresetLock;
        private bool _breastPhysPresetLock;
        private bool _pluginPresetLock;
        private bool _skinPresetLock;
        private bool _morphPresetLock;
        private bool _hairPresetLock;
        private bool _clothingPresetLock;

        public void StorePresetLocks(Atom atom, bool clearAllLocks = false, bool lockClothingPreset = false) {
            var pmControlList = atom.presetManagerControls;
            foreach (var pmc in pmControlList) {
                switch (pmc.name) {
                    case "geometry":
                        _generalPresetLock = pmc.lockParams;
                        break;
                    case "AppearancePresets":
                        _appPresetLock = pmc.lockParams;
                        break;
                    case "PosePresets":
                        _posePresetLock = pmc.lockParams;
                        break;
                    case "AnimationPresets":
                        _animationPresetLock = pmc.lockParams;
                        break;
                    case "FemaleGlutePhysicsPresets":
                        _glutePhysPresetLock = pmc.lockParams;
                        break;
                    case "FemaleBreastPhysicsPresets":
                        _breastPhysPresetLock = pmc.lockParams;
                        break;
                    case "PluginPresets":
                        _pluginPresetLock = pmc.lockParams;
                        break;
                    case "SkinPresets":
                        _skinPresetLock = pmc.lockParams;
                        break;
                    case "MorphPresets":
                        _morphPresetLock = pmc.lockParams;
                        break;
                    case "HairPresets":
                        _hairPresetLock = pmc.lockParams;
                        break;
                    case "ClothingPresets":
                        _clothingPresetLock = pmc.lockParams;
                        break;
                }

                if (pmc.name == "ClothingPresets" && lockClothingPreset) {
                    pmc.lockParams = true;
                } else if (clearAllLocks) {
                    pmc.lockParams = false;
                }
            }
        }

        public void RestorePresetLocks(Atom atom) {
            var pmControlList = atom.presetManagerControls;
            foreach (var pmc in pmControlList) {
                switch (pmc.name) {
                    case "geometry":
                        pmc.lockParams = _generalPresetLock;
                        break;
                    case "AppearancePresets":
                        pmc.lockParams = _appPresetLock;
                        break;
                    case "PosePresets":
                        pmc.lockParams = _posePresetLock;
                        break;
                    case "AnimationPresets":
                        pmc.lockParams = _animationPresetLock;
                        break;
                    case "FemaleGlutePhysicsPresets":
                        pmc.lockParams = _glutePhysPresetLock;
                        break;
                    case "FemaleBreastPhysicsPresets":
                        pmc.lockParams = _breastPhysPresetLock;
                        break;
                    case "PluginPresets":
                        pmc.lockParams = _pluginPresetLock;
                        break;
                    case "SkinPresets":
                        pmc.lockParams = _skinPresetLock;
                        break;
                    case "MorphPresets":
                        pmc.lockParams = _morphPresetLock;
                        break;
                    case "HairPresets":
                        pmc.lockParams = _hairPresetLock;
                        break;
                    case "ClothingPresets":
                        pmc.lockParams = _clothingPresetLock;
                        break;
                }
            }
        }
    }
}
